class Collection(list):
    def __init__(self):
        self.total_count = None
        self.page = None
        self.page_size = None
        super(Collection, self).__init__(self)

    def has_previous(self):
        return self.page is not None and self.page > 1

    def has_next(self):
        return self.page is not None and self.total_count > (self.page * self.page_size)

    def to_dict(self):
        data = self.__dict__
        data['objects'] = self[:]
        data['has_next'] = self.has_next()
        data['has_previous'] = self.has_previous()
        return data



import inspect
import json
from storm.db import Database, ConnectionPool
from storm.error import StormError
from tornado import gen
from tornado.web import RequestHandler
from storm.collection import Collection

class Model(object):
    TYPE_MONGO_DB = 'mongodb'
    TYPE_MYSQL = 'mysql'
    db = None
    check_for_handler = False

    def __init__(self):
        self._type = type(self).__name__.lower()
        self._changes = []

        if not hasattr(self, '_table'):
            self._table = self._type

        if not hasattr(self, '_primary_key'):
            self._primary_key = '_id'

        if not hasattr(self, '_json_fields'):
            self._json_fields = []

    def __setattr__(self, name, value):
        if name[0] == '_':
            self.__dict__[name] = value
            return

        if name in self.__dict__:
            old_value = self.__dict__[name]
            if value != old_value:
                self._changes.append(name)

            self.__dict__[name] = value
            return

        self._changes.append(name)
        self.__dict__[name] = value

    @staticmethod
    def set_db(database):
        if (not isinstance(database, Database) and
            not isinstance(database, ConnectionPool)):

            raise StormError("""database must be instance of storm.db.Database
                or storm.db.ConnectionPool""")

        Model.db = database

        if (not hasattr(Model, '_primary_key') and
            Model.get_database_type() == Model.TYPE_MYSQL):
            Model._primary_key = 'id'

    @staticmethod
    @gen.coroutine
    def get_db(callback=None):
        if Model.check_for_handler:

            # loop over where this came from and if it came from an instance of
            # tornado.web.RequestHandler that has a db property on it then use
            # that.  this isn't really great, but it is a huge convenience to
            # be able to tie a db connection to a request if you want
            for trace in inspect.stack()[3:]:
                local_vars = trace[0].f_locals
                instance = local_vars.get('self', None)
                if instance and isinstance(instance, RequestHandler) and hasattr(instance, 'db'):
                    if callback is not None:
                        callback(instance.db)
                        return

                    raise gen.Return(instance.db)

        if isinstance(Model.db, ConnectionPool):
            db = yield Model.db.get_db()
            if callback is not None:
                callback(db)
                return

            raise gen.Return(db)

        if callback is not None:
            callback(Model.db)
            return

        raise gen.Return(Model.db)

    @classmethod
    def get_table(class_name):
        table = class_name.__name__.lower()
        if hasattr(class_name, '_table'):
            table = getattr(class_name, '_table')

        return table

    @staticmethod
    def get_database_type(db_object=None):
        name = type(Model.db) if db_object is None else type(db_object)
        if db_object is None and isinstance(Model.db, ConnectionPool):
            name = Model.db.get_db_class()

        return name.__name__.lower()

    @gen.coroutine
    def before_save(self):
        pass

    @gen.coroutine
    def after_save(self):
        pass

    @gen.coroutine
    def after_load(self):
        pass

    @classmethod
    def _convert_object(class_name, obj):
        return_obj = class_name()

        # set all properties
        for key in obj:
            if key in return_obj._json_fields:
                try:
                    obj[key] = json.loads(obj[key])
                except:
                    obj[key] = {}

            setattr(return_obj, key, obj[key])

        # make sure the primary key is set to a string
        # for mongodb
        if Model.get_database_type() == Model.TYPE_MONGO_DB:
            setattr(return_obj, return_obj._primary_key,
                    str(getattr(return_obj, return_obj._primary_key)))

        # reset changes
        return_obj._changes = []

        return return_obj

    @classmethod
    @gen.coroutine
    def find_all(class_name, data, **args):
        table = getattr(class_name, 'get_table')()

        callback = None
        if 'callback' in args:
            callback = args['callback']
            del(args['callback'])

        db = yield Model.get_db()
        objects, total_count = yield db.select_multiple(table, data, **args)

        as_dict = args.get('as_dict', False)

        collection = Collection()

        if 'page' in args:
            collection.page = args['page']
            collection.page_size = args.get('page_size', 10)

        collection.total_count = total_count

        for obj in objects:
            new_obj = getattr(class_name, '_convert_object')(obj)
            yield new_obj.after_load()
            collection.append(new_obj if not as_dict else new_obj.__dict__)

        if callback is None:
            raise gen.Return(collection)

        callback(collection)

    @classmethod
    @gen.coroutine
    def find(class_name, **args):
        table = getattr(class_name, 'get_table')()

        callback = None
        if 'callback' in args:
            callback = args['callback']
            del(args['callback'])

        db = yield Model.get_db()
        obj = yield db.select_one(table, **args)

        return_obj = None
        if obj is not None:
            return_obj = getattr(class_name, '_convert_object')(obj)
            yield return_obj.after_load()

        if callback is None:
            raise gen.Return(return_obj)

        callback(return_obj)

    @gen.coroutine
    def save(self, callback=None):
        yield self.before_save(self._changes)

        to_save = {}
        for k in [key for key in self.__dict__ if not key[0] == '_']:
            val = self.__dict__[k]
            if k in self._json_fields:
                try:
                    val = json.dumps(val)
                except:
                    pass

            to_save[k] = val

        is_compound_primary_key = isinstance(self._primary_key, list)
        primary_key_not_included = not is_compound_primary_key and not hasattr(self, self._primary_key)
        primary_key_was_set = self._primary_key in self._changes

        if is_compound_primary_key:
            primary_key_was_set = True
            for field in self._primary_key:
                if field not in self._changes:
                    primary_key_was_set = False
                    break

        if primary_key_not_included or primary_key_was_set:
            db = yield Model.get_db()
            result = yield db.insert(self._table, to_save)

            if Model.get_database_type() == Model.TYPE_MONGO_DB:
                result = str(result)

            if primary_key_not_included:
                setattr(self, self._primary_key, result)
        else:

            # I do not remember why this is here, but I think it might be for
            # mongodb where the primary key starts with an underscore and
            # therefore won't be included in the to_save dictionary
            if not is_compound_primary_key:
                to_save[self._primary_key] = self.__dict__[self._primary_key]

            db = yield Model.get_db()
            result = yield db.update(self._table, to_save, self._changes, self._primary_key)

        yield self.after_save(self._changes)
        self._changes = []

        if callback is None:
            raise gen.Return(result)

        callback(result)

    @gen.coroutine
    def delete(self, callback=None):
        result = False

        is_compound_primary_key = isinstance(self._primary_key, list)
        if is_compound_primary_key or hasattr(self, self._primary_key):
            primary_key_fields = self._primary_key
            primary_key_values = []

            if not is_compound_primary_key:
                primary_key_fields = [primary_key_fields]

            for key in primary_key_fields:
                primary_key_values.append(getattr(self, key))

            db = yield Model.get_db()
            result = yield db.delete(self._table,
                                     primary_key_fields,
                                     primary_key_values)

        if callback is None:
            raise gen.Return(result)

        callback(result)



import motor
from tornado import gen
from bson.objectid import ObjectId
from storm.db import Database, ConnectionPool
from storm import error


class MongoDb(Database):
    def connect(self):
        if self.is_connected:
            return

        self.motor_client = motor.MotorClient(
            self.connection.host,
            self.connection.port
        ).open_sync()

        self.db = self.motor_client[self.connection.database]
        self.is_connected = True

    @gen.coroutine
    def select_one(self, table, **kwargs):
        self.connect()

        if '_id' in kwargs:
            kwargs['_id'] = ObjectId(kwargs['_id'])

        result = yield motor.Op(getattr(self.db, table).find_one, kwargs)

        if result is None:
            raise error.StormNotFoundError("Object of type: %s not found with args: %s" % (table, kwargs))

        callback = kwargs.get('callback')

        if callback is None:
            raise gen.Return(result)

        callback(result)

    @gen.coroutine
    def select_multiple(self, table, data, **kwargs):
        cursor = getattr(self.db, table).find(data)

        # handle pagination
        if 'page' in kwargs:
            page = kwargs['page']
            page_size = kwargs.get('page_size', 10)
            offset = (page - 1) * page_size
            cursor.limit(page_size)
            cursor.skip(offset)

        # default sort to newest first
        sort = kwargs.get('sort', [('_id', 0)])
        cursor.sort(sort)

        data = []
        while (yield cursor.fetch_next):
            data.append(cursor.next_object())

        total_count = yield motor.Op(cursor.count)

        callback = kwargs.get('callback')

        if callback is None:
            raise gen.Return([data, total_count])

        callback([data, total_count])

    @gen.coroutine
    def insert(self, table, data, callback=None):
        result = yield self.update(table, data, callback)

        if callback is None:
            raise gen.Return(result)

        callback(result)

    @gen.coroutine
    def update(self, table, data, changes, primary_key, callback=None):
        self.connect()

        if primary_key in data:
            data[primary_key] = ObjectId(data[primary_key])

        result = yield motor.Op(self.db[table].save, data)

        if callback is None:
            raise gen.Return(result)

        callback(result)

    @gen.coroutine
    def delete(self, table, primary_key_fields, primary_key_values, callback=None):
        self.connect()

        to_delete = {
            primary_key_fields[0]: ObjectId(primary_key_values[0])
        }

        result = yield motor.Op(self.db[table].remove, to_delete)

        if callback is None:
            raise gen.Return(result)

        callback(result)


class ConnectionPool(ConnectionPool):
    def get_db_class(self):
        return MongoDb



import sys
import datetime
from tornado import gen
from storm.db import Database, ConnectionPool
from storm import error
import tornado_mysql
from tornado_mysql.pools import Pool

cursor_type = tornado_mysql.cursors.DictCursor


class MySql(Database):
    @gen.coroutine
    def connect(self, callback=None):
        if not self.is_connected:
            self.db = Pool(
                dict(user=self.connection.user,
                     passwd=self.connection.password,
                     db=self.connection.database,
                     cursorclass=cursor_type),
                max_idle_connections=1,
                max_open_connections=1)

            self.is_connected = True

        if callback is None:
            raise gen.Return(True)

        callback(True)

    @gen.coroutine
    def close(self, callback=None):
        if self.is_connected:
            yield self.db.close()
            self.is_connected = False

        if callback is None:
            raise gen.Return(True)

        callback(True)

    @staticmethod
    def _quote(value):
        if value is None:
            return 'null'

        if isinstance(value, datetime.datetime):
            return "'%s'" % str(value)

        if value == 'NOW()':
            return value

        # convert python3 byte strings
        if sys.version_info >= (3,0,0) and isinstance(value, bytes):
            value = value.decode('utf-8')

        if isinstance(value, float):
            return str(value)

        try:
            value = str(int(value))
        except:
            value = "'%s'" % tornado_mysql.converters.escape_string(value)

        return value

    @gen.coroutine
    def select_one(self, table, **kwargs):
        yield self.connect()

        where_bits = []
        for key in kwargs:
            where_bits.append("`%s` = %s" % (key, MySql._quote(kwargs[key])))

        sql = "SELECT * FROM `%s` WHERE BINARY %s" % (table, ' AND BINARY '.join(where_bits))

        cur = yield self.db.execute(sql)
        result = cur.fetchone()

        if result is None:
            raise error.StormNotFoundError("Object of type: %s not found with args: %s" % (table, kwargs))

        callback = kwargs.get('callback')
        if callback is None:
            raise gen.Return(result)

        callback(result)


    @gen.coroutine
    def select_multiple(self, table, query, **kwargs):
        yield self.connect()

        query.bind(':table', table)

        page = kwargs.get('page')
        if page:
            page_size = kwargs.get('page_size', 10)
            query.limit = page_size
            query.offset = (page - 1) * page_size

        raw_sql = query.sql

        total_count = 0

        tasks = [self.db.execute(raw_sql)]
        if page:
            tasks.append(self.db.execute(query.count_sql))

        cursors = yield tasks

        results = [cursors[0].fetchall()]
        if len(cursors) > 1:
            results.append(cursors[1].fetchall())

        data = results[0]
        total_count = len(data)
        if len(results) == 2:
            total_count = results[1][0]['count']

        data, filtered_out_count = query.apply_filters(data)
        total_count -= filtered_out_count

        callback = kwargs.get('callback', None)
        if callback is None:
            raise gen.Return([data, total_count])

        callback([data, total_count])


    @gen.coroutine
    def insert(self, table, data, callback=None):
        yield self.connect()

        fields = []
        values = []
        for key in data:
            fields.append(key)
            value = MySql._quote(data[key])
            values.append(value)

        sql = "INSERT INTO `%s` (`%s`) VALUES (%s)" % (table, '`, `'.join(fields), ', '.join(values))

        # double escape % sign so we don't get an error if one of the fields
        # we are trying to insert has a % in it
        sql = sql.replace('%', '%%')

        cur = yield self.db.execute(sql)
        insert_id = cur.lastrowid

        if callback is None:
            raise gen.Return(insert_id)

        callback(insert_id)

    @gen.coroutine
    def update(self, table, data, changes, primary_key, callback=None):
        if len(changes) == 0:
            raise gen.Return(False)

        yield self.connect()

        if 'modified_on' in data:
            changes.append('modified_on')
            data['modified_on'] = 'NOW()'

        pairs = []
        compound_primary_key = isinstance(primary_key, list)
        for key in changes:
            if compound_primary_key and key in primary_key:
                continue

            if key == primary_key:
                continue

            pairs.append("`%s` = %s" % (key, MySql._quote(data[key])))

        if not compound_primary_key:
            primary_key = [primary_key]

        where_bits = []
        for key in primary_key:
            where_bits.append("`%s` = %s" % (key, MySql._quote(data[key])))

        sql = "UPDATE `%s` SET %s WHERE %s" % (table, ', '.join(pairs), ' AND '.join(where_bits))

        result = yield self.db.execute(sql)
        if callback is None:
            raise gen.Return(result)

        callback(result)

    @gen.coroutine
    def delete(self, table, primary_key_fields, primary_key_values, callback=None):
        self.connect()

        where_bits = []
        for i, key in enumerate(primary_key_fields):
            where_bits.append("`%s` = %s" % (key, MySql._quote(primary_key_values[i])))

        result = False
        if len(where_bits) > 0:
            sql = "DELETE FROM `%s` WHERE %s" % (table, ' AND '.join(where_bits))
            result = yield self.db.execute(sql)

        if callback is None:
            raise gen.Return(result)

        callback(result)

class QueryFilter(object):
    TYPE_EQUAL = '='
    TYPE_NOT_EQUAL = '!='
    TYPE_GREATER_THAN = '>'
    TYPE_GREATER_THAN_OR_EQUAL = '>='
    TYPE_LESS_THAN = '<'
    TYPE_LESS_THAN_OR_EQUAL = '<='
    TYPE_IN = 'in'
    TYPE_NOT_IN = 'not in'

    def __init__(self, key, comparison, value):
        self.key = key
        self.comparison = comparison.lower()
        self.value = value

    def matches(self, row):
        if self.comparison == self.TYPE_EQUAL:
            return row[self.key] == self.value
        elif self.comparison == self.TYPE_NOT_EQUAL:
            return row[self.key] != self.value
        elif self.comparison == self.TYPE_GREATER_THAN:
            return row[self.key] > self.value
        elif self.comparison == self.TYPE_GREATER_THAN_OR_EQUAL:
            return row[self.key] >= self.value
        elif self.comparison == self.TYPE_LESS_THAN:
            return row[self.key] < self.value
        elif self.comparison == self.TYPE_LESS_THAN_OR_EQUAL:
            return row[self.key] <= self.value
        elif self.comparison == self.TYPE_IN:
            return row[self.key] in self.value
        elif self.comparison == self.TYPE_NOT_IN:
            return row[self.key] not in self.value

        return True

class Query(object):
    def __init__(self, sql):
        self._sql = sql
        self.count_sql = None
        self.to_bind = {}
        self.limit = None
        self.offset = None
        self.filters = []

    def bind(self, key, value):
        self.to_bind[key] = value
        return self

    def filter(self, key, comparison, value):
        self.filters.append(QueryFilter(key, comparison, value))
        return self

    def all_filters_allow(self, row):
        for f in self.filters:
            if not f.matches(row):
                return False

        return True

    def apply_filters(self, data):
        if self.limit or self.offset:
            raise error.StormError("""You cannot apply filters when using page
                                   and page_size to limit the mysql data""")

        if len(data) == 0 or len(self.filters) == 0:
            return (data, 0)

        new_data = []
        num_removed = 0
        for row in data:
            if self.all_filters_allow(row):
                new_data.append(row)
                continue

            num_removed += 1

        return (new_data, num_removed)

    @property
    def sql(self):
        sql = self._sql
        if ':table' in self.to_bind:
            sql = sql.replace(':table', "`%s`" % self.to_bind[':table'])
            del(self.to_bind[':table'])

        for key in self.to_bind:
            sql = sql.replace(key, MySql._quote(self.to_bind[key]))

        if self.limit:
            self.count_sql = "SELECT count(*) count FROM %s" % sql.split(' FROM ')[1]

        if self.limit:
            sql += " LIMIT %d" % self.limit

        if self.offset:
            sql += " OFFSET %d" % self.offset

        return sql


class ConnectionPool(ConnectionPool):
    def __init__(self, connection, count=10, lifetime=3600):
        super(ConnectionPool, self).__init__(connection, count, lifetime)
        db = MySql(self.connection)
        db.db = Pool(
            dict(user=connection.user,
                 passwd=connection.password,
                 db=connection.database,
                 cursorclass=cursor_type),
            max_idle_connections=self.count,
            max_recycle_sec=self.lifetime,
            max_open_connections=self.count+10)

        db.is_connected = True
        self._db = db

    def get_db_class(self):
        return MySql

    @gen.coroutine
    def get_db(self, callback=None):
        if callback is not None:
            callback(self._db)
            return

        raise gen.Return(self._db)



import time
from storm import error
from tornado import gen


class Connection(object):
    def __init__(self, host='localhost', port=None, database=None, user=None, password=None):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password


class ConnectionPool(object):
    def __init__(self, connection, count=10, lifetime=3600):
        self.connection = connection
        self.count = count
        self.lifetime = lifetime

    @gen.coroutine
    def get_db(self, callback=None):
        raise NotImplementedError('The "get_db" method is not implemented')

    def get_db_class(self):
        raise NotImplementedError('The "get_db_class" method is not implemented')


class Database(object):
    def __init__(self, connection):
        if not isinstance(connection, Connection):
            raise error.StormError('connection must be instance of storm.db.Connection')

        self.connection = connection
        self.is_connected = False
        self.start_time = time.time()



from setuptools import setup, find_packages
from storm import version

# magic
setup(
    name='tornado-storm',
    version=version,
    description='A simple ORM for Tornado',
    author='Craig Campbell',
    author_email='iamcraigcampbell@gmail.com',
    url='https://github.com/ccampbell/storm',
    download_url='https://github.com/ccampbell/storm/archive/%s.zip#egg=tornado-storm-%s' % (version, version),
    license='MIT',
    install_requires=['tornado >= 4.0', 'tornado_mysql >= 0.5', 'motor >= 0.1.1'],
    packages=find_packages(),
    py_modules=['storm'],
    platforms=["any"]
)
